/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: PISymTab.cpp$
 * $Date: Sun Aug 10 06:29:55 1997$
 *
 Description:
\*____________________________________________________________________________*/
//$SourceTop:$

#include <assert.h>
#include <ctype.h>

#include "FastLib.h"
#include "DblList.h"
#include "Utils.h"
#include "PIDefs.h"
#include "PiAPI.h"

#if defined(NDEBUG)
#	define MUSTBETRUE(x)	(x)
#else
#	define MUSTBETRUE(x)	(assert(x))
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class PISymbolTable
{
private:
	Semaphore *pMutex;
	enum { HASH_TABLE_SIZE=1007 };
	DblList aList[HASH_TABLE_SIZE];

	/* --- forbid copy constructor --- */
	PISymbolTable( const PISymbolTable & )
	:	pMutex( 0 )
		{ assert( 0 ); };

public:
	/*
	*/
	PISymbolTable()
		:	pMutex( PIPlatform_allocLocalMutex() )
		{
		MUSTBETRUE( pMutex );
		};

	/*
	*/
	~PISymbolTable()
		{
		MUSTBETRUE( pMutex );
		MUSTBETRUE( !PISync_lock( pMutex ) );
		for( int i=0; i<HASH_TABLE_SIZE; i++ )
			{
			DblList &tList = aList[i];
			while( tList.Size() )
				{
				PI_DELETE( [] tList.Cap() );
				};
			};
		MUSTBETRUE( !PISync_unlock( pMutex ) );
		MUSTBETRUE( !PISync_delete( pMutex ) );
		};
	
	/*
	*/
	inline static char nop( char c ) { return c; };
	inline static char tolc( char c ) { return (tolower(c)); };
	inline int HashIndex( const char *pS, int iCaseSensitive )
		{
		unsigned uH = 0;
		if ( iCaseSensitive )
			{ return FastLib_getHashKey( pS, HASH_TABLE_SIZE ); }
		else
			{ return FastLib_getHashKeyI( pS, HASH_TABLE_SIZE ); };
		};

	/*
	*/
	inline DblList &HashLookupList( const char *pS, int iCaseSensitive )
		{
		assert( pS );
		return aList[ HashIndex(pS, iCaseSensitive ) ];
		};

	/*
	*/
	inline const char *UnsafeInsert( const char *pS, int iCaseSensitive )
		{
		/* --- this function assumes the element does not already
		exist in the list --- */
		assert( !UnsafeLookup( pS, iCaseSensitive ) );
		DblList &tList = HashLookupList( pS, iCaseSensitive );
		char *pNewS = DuplicateString( pS );
		tList.Append( (DblList::type)pNewS );
		return pNewS;
		};

	/*
	*/
	inline const char *Insert( const char *pS, int iCaseSensitive )
		{
		assert( pS );
		if ( !pS ) { return 0; };
		MUSTBETRUE( !pMutex );
		MUSTBETRUE( !PISync_lock( pMutex ) );
		const char *pTmp = UnsafeInsert( pS, iCaseSensitive );
		MUSTBETRUE( !PISync_unlock( pMutex ) );
		return pTmp;
		};

	/*
	*/
	inline const char *UnsafeLookup( const char *pS, int iCaseSensitive )
		{
		DblList &tList = HashLookupList( pS, iCaseSensitive );
		for( DblListIterator i(tList); !i.BadIndex(); i++)
			{
			if ( iCaseSensitive )
				{
				if ( !strcmp( (const char *)i.Current(), pS ) )
					{ return (const char *)i.Current(); };
				}
			else
				{
				if ( !PIUtil_stricmp( (const char *)i.Current(), pS ) )
					{ return (const char *)i.Current(); };
				}
			};
		return 0;
		};

	/*
	*/
	inline const char *Lookup( const char *pS, int iCaseSensitive )
		{
		assert( pS );
		if ( !pS ) { return 0; };
		MUSTBETRUE( pMutex );
		MUSTBETRUE( !PISync_lock( pMutex ) );
		const char *pTmp = UnsafeLookup( pS, iCaseSensitive );
		MUSTBETRUE( !PISync_unlock( pMutex ) );
		return pTmp;
		};

	/*
	*/
	inline const char *LookupOrInsert( const char *pS, int iCaseSensitive )
		{
		assert( pS );
		if ( !pS ) { return 0; };
		MUSTBETRUE( pMutex );
		MUSTBETRUE( !PISync_lock( pMutex ) );
		const char *pTmp = UnsafeLookup( pS, iCaseSensitive );
		if ( !pTmp )
			{
			pTmp = UnsafeInsert( pS, iCaseSensitive );
			};
		MUSTBETRUE( !PISync_unlock( pMutex ) );
		return pTmp;
		};
};

/*____________________________________________________________________________*\
 *
 Global declarations:
	singleton instance of the symbol table
\*____________________________________________________________________________*/
static PISymbolTable *pTable=0;

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void PISymbolTable_Init()
{
	assert( !pTable );
	pTable = PI_NEW( PISymbolTable() );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void PISymbolTable_Cleanup()
{
	assert( pTable );
	PI_DELETE( pTable );
	pTable = 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
const char *PISymbolTable_Insert( const char *pS, int iCaseSensitive )
{
	assert( pTable );
	return pTable->Insert( pS, iCaseSensitive );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
const char *PISymbolTable_Lookup( const char *pS, int iCaseSensitive )
{
	assert( pTable );
	return pTable->Lookup( pS, iCaseSensitive );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
const char *PISymbolTable_LookupOrInsert( const char *pS, int iCaseSensitive )
{
	assert( pTable );
	return pTable->LookupOrInsert( pS, iCaseSensitive );
}

