/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: PathMap.cpp$
 * $Date: Sun Aug 10 06:36:50 1997$
 *
 Description:
	Path mapping.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <assert.h>

#include "DblList.h"		/* --- from Base library --- */

#include "HandBase.h"
#include "PIHTTP.h"
#include "PIStrStr.h"
#include "Pi3Expr.h"
#include "HTTPDefs.h"
#include "HTTPCore.h"

/*____________________________________________________________________________*\ *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_FROM			"From"
#define KEY_CONF_TO				"To"
#define KEY_CONF_CASESENSITIVE	"CaseSensitive"
#define KEY_CONF_PATHINFO		"PathInfo"
#define KEY_CONF_ACTION			"Action"
#define VALUE_NO				"No"
#define VALUE_YES				"Yes"

#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	PathMapper

Description:
	Maps a virtual path to another virtual path or a physical path.

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>From
<TD>+
<TD>&ltpath&gt
<TD>Search path
<TD>From="/"

<TR>
<TD>To
<TD>+
<TD>&ltpath&gt
<TD>Replace path
<TD>To="/home/WWWDocs/"

<TR>
<TD>CaseSensitive
<TD>Yes
<TD>Yes|No
<TD>Specifies case sensitivity of "From" matching
<TD>CaseSensitive="No"

<TR>
<TD>PathInfo
<TD>No
<TD>Yes|No
<TD>Specifies if path after "To" is PATH_INFO
<TD>PathInfo="Yes"

<TR>
<TD>Action
<TD>-
<TD>A Pi3 Expression
<TD>Expression to be evaluated if path is mapped
<TD>Action="&dbreplace(response,string,ObjectMap,Scripts)"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>
	From
</H5>
Specifies the virtual path to be replaced. The first and last characters
of this virtual path must be '/'.
<H5>
	To
</H5>
Specifies the path to replace the entire virtual path.
<H5>
	CaseSensitive
</H5>
Specifies whether or not matching the "From" pattern to the virtual
path will be case sensitive or not.
<H5>
	PathInfo
</H5>
Specifies whether or not the pattern which follows the "From" section
of the virtual path will be appended to the "To" path or placed in the
PATH_INFO variable.<P>Typically CGI mapping sets PATH_INFO whereas regular
document mapping does not.
<H5>
	Action
</H5>
Specifies a Pi3 expression which will be evaluated if the path is mapped by
this handler. This allows a 'marker' to be set to effect the processing of 
later phases. This directive may be specified multiple times for multiple
different actions.

Phase:
	MAPPING

Returns:
	PIAPI_ERROR if this handler was invoked for an inappropriate phase, 
	otherwise, PIAPI_COMPLETED if a path was mapped and PIAPI_CONTINUE if
	no path was mapped.
		
Note:
	The 'From' pattern must begin and end with '/' character, otherwise an
	initialization error will be raised. This is to ensure that path mapping
	does not result in unintended files becoming visible.
	<P>
	Note that paths are mapped in the order listed in the configuration file,
	so a path mapping statement with From="/" would match before one with
	From="/icons/" is listed first. Path mapping should be list in reverse
	order of directory depth to ensure correct mapping (place mapper with
	From="/" AFTER the one with From="/icons/").

Example:
	<PRE>
	&ltObject&gt
		Name PathMapper
		Class PathMapperClass
	&lt/Object&gt

	&ltObject&gt
		...
		CheckAccess PathMapper From="/" To="/home/Pi3Web/WebRoot/"
		...
	&lt/Object&gt

	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class PathMapper : public HandlerBaseHTTP
{
private:
	PIString sFrom;
	PIString sTo;
	int iCaseSensitive;
	int iPathInfo;
	DblList lActions;			/* allow multiple actions */
	const char *pFKPath;		/* faster access to 'Path' variable */

protected:
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "PathMapper: ";
		if ( !PIUtil_stricmp( KEY_CONF_FROM, pVariable ) )
			{
			sFrom = pValue;

            /* ---
            Bad things can happen wrt to server compromise if the
            specified URL path does not begin and end with '/'
            --- */
            const char *pTmp = sFrom;
            if ( *pTmp!=HTTP_DIRSEPERATOR &&
                 pTmp[sFrom.Len()-1]!=HTTP_DIRSEPERATOR )
                {
                os << "'From' value must begin and end with a '/'." << ends;
                CONFIG_ERR( Object(), os.str() );
                return 0;
                };
			}
		else if ( !PIUtil_stricmp( KEY_CONF_TO, pVariable ) )
			{
			sTo = pValue;
            Pi3String *pTmp = Pi3String_new( pValue );
            HTTPCore_relativeToAbsolutePath( PIObject_getDB( Object() ),
                pValue, pTmp );
            sTo = Pi3String_getPtr( pTmp );
            Pi3String_delete( pTmp );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_CASESENSITIVE, pVariable ) )
			{
			if ( !PIUtil_stricmp( VALUE_NO, pValue ) )
				{ iCaseSensitive = 0; }; 
			}
		else if ( !PIUtil_stricmp( KEY_CONF_PATHINFO, pVariable ) )
			{
			if ( !PIUtil_stricmp( VALUE_YES, pValue ) )
				{ iPathInfo = 1; }; 
			}
		else if ( !PIUtil_stricmp( KEY_CONF_ACTION, pVariable ) )
			{
			Pi3Expression *pExpr = 0;
			Pi3String *pError = Pi3String_new( 0 );
			pExpr = Pi3Expression_new( pValue, 0, pError );
			if ( !pExpr )
				{
				os << "Bad 'Action' String: " << Pi3String_getPtr( pError )
					<< ends;
				Pi3String_delete( pError );
				CONFIG_ERR( Object(), os.str() );
				Pi3Expression_delete( pExpr );
				pExpr = 0;
				return 0;
				};
			Pi3String_delete( pError );
			lActions.Append( (DblList::type) pExpr );
			}
		else
			{
			os << "Unknown directive '" << pVariable << "'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};	

		return 1;
		};

public:
	PathMapper( PIObject *pTheObject, int iArgc, const char *ppArgv[] )
	:	HandlerBaseHTTP( pTheObject ),
		iCaseSensitive( 1 ),
		iPathInfo( 0 ),
		pFKPath( PIDB_getFastKey( KEY_INT_PATH, PIDBTYPE_STRING ) )
		{	
		ReadParameters( iArgc, ppArgv );

		/* ---
		Validate parameters
		--- */
		if ( sFrom==PIString::Empty() )
			{
			CONFIG_ERR( Object(), "PathMapper: 'From' not defined" );
			};
		if ( sTo==PIString::Empty() )
			{
			CONFIG_ERR( Object(), "PathMapper: 'To' not defined" );
			};
		};

	virtual ~PathMapper()
		{
		for ( DblListIterator i( lActions ); !i.BadIndex(); i++ )
			{
			Pi3Expression_delete( (Pi3Expression *)i.Current() );
			};
		};

	/* -------------- +++++++++++++++++++++++++++++++ ------------------ *\

								Utilities

	\* -------------- +++++++++++++++++++++++++++++++ ------------------ */
	/* ---
	Create a translated PATH_INFO
	--- */
	void CreatePathTranslated( PIHTTP &tPIHTTP )
		{
		PIDB *pR = tPIHTTP.pResponseDB;
		const char *pPathInfo = (const char*)PIDB_lookup( pR, PIDBTYPE_STRING, 
			KEY_INT_PATHINFO, 0 );
		if ( pPathInfo )
			{
			/*
			** Path translated is the PATH_INFO, mapped from virtual to
			** physical path
			*/
			/* --- make sub request context --- */
			PIHTTP *pChildHTTP = PIHTTP_newChild( &tPIHTTP );

			/* --- set path --- */
			PIDB_replace( pChildHTTP->pResponseDB, PIDBTYPE_STRING,
				KEY_INT_PATH, (void *)pPathInfo, 0 );

			/* --- dispatch the sub request across the mapping phase --- */
			int iRet = PIAPI_ERROR;
			iRet = HTTPCore_dispatch(
				pChildHTTP,
				PH_MAPPING,
				PH_MAPPING );
			
			/* --- copy the childs path to this path translated --- */
			if ( iRet==PIAPI_COMPLETED &&
				((pChildHTTP->iStatus==0) || (pChildHTTP->iStatus==ST_OK)) )
				{
				void *pPath = PIDB_lookup( pChildHTTP->pResponseDB,
					PIDBTYPE_STRING, KEY_INT_PATH, 0 );
				PIDB_replace( pR, PIDBTYPE_STRING, KEY_INT_PATHTRANSLATED, 
					pPath, 0 );
				}
			else
				{
				PIDB_replace( pR, PIDBTYPE_STRING, KEY_INT_PATHTRANSLATED, 
					0, 0 );
				};
			PIHTTP_delete( pChildHTTP );
			};
		};

	/* ---
	Create script name variable. The is the URI without the
	path info at the end.
	--- */
	void CreateScriptName( PIDB *pQ, PIDB *pR )
		{
		const char *pURI = (const char*)PIDB_lookup( pQ, PIDBTYPE_STRING, 
			KEY_HTTP_URI, 0 );
		const char *pPathInfo = (const char*)PIDB_lookup( pR, PIDBTYPE_STRING, 
			KEY_INT_PATHINFO, 0 );
		
		if ( !pURI )
			{ return; };
		if ( !pPathInfo )
			{
			PIDB_add( pR, PIDBTYPE_STRING, KEY_INT_SCRIPTNAME,
				(void *)pURI, 0 );
			return;
			};
		int iULen = strlen( pURI );
		int iPLen = strlen( pPathInfo );
		if ( iULen < iPLen )
			{ return; };
		PIString sTmp( pURI, iULen - iPLen );
		PIDB_add( pR, PIDBTYPE_STRING, KEY_INT_SCRIPTNAME,
			(void *)(const char *)sTmp, 0 );
		};

	/* -------------- +++++++++++++++++++++++++++++++ ------------------ *\
	
								Real stuff

	\* -------------- +++++++++++++++++++++++++++++++ ------------------ */
	int Handle( int /* iPhase */, PIHTTP &tPIHTTP, PIIOBuffer & )
		{
		PIDB *pR = tPIHTTP.pResponseDB;
		PIDB *pQ = tPIHTTP.pRequestDB;

		/* ---
		Check if the first component of the path matches
		--- */
		/* --- get Path --- */
		const char *pPath = (const char *)PIDB_lookup( pR, PIDBTYPE_STRING,
			pFKPath, PIDBFLAG_FASTKEY );
		if ( !pPath )
			{ return PIAPI_ERROR; };

		int iMatches = iCaseSensitive ?
			!strncmp( pPath, sFrom, sFrom.Len() ):
			!PIUtil_strncmpi( pPath, sFrom, sFrom.Len() );
		PIString sNewPath( sTo );
		if ( !iMatches )
			{
			return PIAPI_CONTINUE;
			};

		const char *pPathInfo = 0;
		if ( iPathInfo )
			{
			const char *pTmp = &( pPath[sFrom.Len()] );
			int i;
			for( i=0; pTmp[i]; i++ )
				{
				if ( pTmp[i]==HTTP_DIRSEPERATOR )
					{
					/*
					** NOTE: maybe review this
					*/
					pPathInfo = &( pTmp[i] );
					break;
					};
				};
			if ( pPathInfo )
				{
				PIDB_add( pR, PIDBTYPE_STRING, KEY_INT_PATHINFO,
					(void *)pPathInfo, 0 );
				};
			sNewPath.Concatenate( PIString( pTmp, i ) );
			}
		else
			{
			sNewPath.Concatenate( &( pPath[sFrom.Len()] ) );
			};

		PIDB_replace( pR, PIDBTYPE_STRING, KEY_INT_PATH,
            (void *)(const char *)sNewPath, 0 );

		/* ---
		Setup path translated and script name
		--- */
		CreatePathTranslated( tPIHTTP );
		CreateScriptName( pQ, pR );

		/* ---
		Evaluate success expressions
		--- */
		for( DblListIterator i( lActions ); !i.BadIndex(); i++ )
			{
			Pi3Expression_write( (Pi3Expression *)i.Current(),
				&tPIHTTP, 0, 0, 0 );
			};

		return PIAPI_COMPLETED;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PathMapper_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseHTTP_constructor(pObj, new PathMapper( pObj, iArgc, ppArgv));
}
#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name PathMapperClass
		Type LogicExtension
		Library HTTP
		OnClassLoad HandlerBaseHTTP_onClassLoad
		Constructor PathMapper_constructor
		CopyConstructor HandlerBaseHTTP_copyConstructor
		Destructor HandlerBaseHTTP_destructor
		Execute HandlerBaseHTTP_execute
	</Class>

	<Object>
		Name PathMapper
		Class PathMapperClass
	</Object>

/*___+++CNF_END+++___*/
#endif

