/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: StubIO.cpp$
 * $Date: Sun Aug 10 06:35:52 1997$
 *
 Description:
	Stubs out an IO layer, by 'faking' real new connections on
	a configurable interval. New connections are then fed canned
	data from a file. Output data is discarded.
\*____________________________________________________________________________*/
/* $SourceTop:$ */

#include <assert.h>
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "Pi2API.h"
#include "DblList.h"		/* from Base library */
#include "PIStrStr.h"		/* from Base library */

/* --- Values from configuration file --- */
#define KEY_CONF_SENDINTERVAL	"SendInterval"
#define KEY_CONF_ACCEPTINTERVAL	"AcceptInterval"
#define KEY_CONF_DATAFILE		"DataFile"
#define KEY_CONF_HTTPEQUIV		"HTTPEquiv"
#define KEY_CONF_ERRORAFTER		"ErrorAfter"

/* --- other macros --- */
#define CONFIG_ERR(this, msg)   \
	{ PILog_addMessage( PIObject_getDB(this), \
	PIObject_getConfigurationDB(this), PILOG_ERROR, (msg) ); }

/*
#define D { cerr << __FILE__ << ": " << __LINE__ << endl; };
*/
#define D

#define DEFAULT_INTERVAL	(0)

/*____________________________________________________________________________*\
 Description:
	Documentation for this IO object.
\*____________________________________________________________________________*/

#if 0
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	StubIO

Description:
	This object 'fakes' an IO connection by feeding canned data to the
	reader. The canned data is read from files.

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>AcceptInterval
<TD>0
<TD>0, -1, n
<TD>IO accept interval
<TD>AcceptInterval 0

<TR>
<TD>SendInterval
<TD>0
<TD>0, -1, n
<TD>IO output interval
<TD>SendInterval 0

<TR>
<TD>Data
<TD>+
<TD>&ltfilename&gt
<TD>File data to read
<TD>DataFile ./../tests/t2.tst

#define KEY_CONF_ERRORAFTER		"ErrorAfter"
<TR>
<TD>HTTPEquiv
<TD>+
<TD>&ltvariable&gt: &ltvalue&gt
<TD>Variable value pair for DB
<TD>HTTPEquiv="RemoteAddr: 127.0.0.1"

<TR>
<TD>ErrorAfter
<TD>+
<TD>&ltnumber&gt
<TD>Number of child IO objects to error after
<TD>ErrorAfter="10000"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>AcceptInterval</H5>
Specify an interval measured in seconds to wait everytime a new connection
is generated by copying the prototype IO object.

<H5>SendInterval</H5>
Specify an interval measured in seconds to wait everytime data is sent. 

<H5>DataFile</H5>
Specifies a file with data to return to the reader. The file path 
should be absolute or relative to the current directory. This parameter
can be specified multiple times to specify different files that will
be assigned to different readers in cycles.

<H5>HTTPEquiv</H5>
Specifies an RFC822 variable value pair to put in this objects DB.

<H5>ErrorAfter</H5>
Specify the number of IO error objects to to error after. This facilites
can be used to test the server for a discrete number of IO requests before
exiting.

Returns:
	PIAPI_COMPLETED on success.
	PIAPI_ERROR and PIAPI_ABORT respectively for generic and severe
	error conditions.

Example:
	<PRE>
	&ltObject&gt
		Name StubIO
		Class StubIOClass
		AcceptInterval 0
		SendInterval 0
		DataFile ./../tests/t2.tst
	&lt/Object&gt
	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Parse, store and add RFC822 headers to  response DB's.
\*____________________________________________________________________________*/
class RFC822Header
{
private:
	const char *pFKVariable;
	PIString sValue;
	int iOK;

public:
	RFC822Header( const char *pLine, int iType )
	:   pFKVariable( 0 ),
		iOK( 0 )
		{
		if ( !pLine )
			{ return; };
		int iLen = strlen( pLine );
		int i;
		 for( i=0; i<iLen && (isalnum(pLine[i])); i++);
		if ( !i || pLine[i]!=':' )
			{ return; };
		PIString sVariable( pLine, i );
		for( i++; i<iLen && (isspace(pLine[i])); i++);  /* scan past ws */
		pFKVariable = PIDB_getFastKey( sVariable, iType );
		sValue = &( pLine[i] );
		iOK = 1;
		};

	/* --- Replace this header in DB --- */
	void ReplaceHeader( PIDB *pDB, int iType )
		{
		assert( pDB );
		assert( pFKVariable );
		assert( iOK );
		PIDB_replace( pDB, iType, pFKVariable,
			(void *)(const char *)sValue, PIDBFLAG_FASTKEY );
		};
	inline int IsOK()
		{ return iOK; };
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class IOResponseRecord
{
public:
	char *pInData;		/* this is the input data to the IO stream */
	int iDataSize;		/* size of pInData */

	IOResponseRecord( char *pTheInData, int iTheDataSize )
	: pInData( pTheInData ), iDataSize( iTheDataSize )
		{};

	~IOResponseRecord()
		{
		PI_DELETE( [] pInData );
		};

	static IOResponseRecord *MakeResponseRecord( const PIObject *pObject,
		const char *pDataFile )
		{
		assert( pObject && pDataFile );

		PIFInfo *pInfo = PIFInfo_new( pDataFile );
		ifstream ifs;
		if ( pInfo )
			{ ifs.open( pDataFile ); };
		if ( !pInfo || !ifs.good() )
			{
			PILog_addMessage( PIObject_getDB(pObject),
				PIObject_getConfigurationDB(pObject), PILOG_ERROR,
				"StubIO: Datafile, '%s' could not be opened.", pDataFile );
			return 0;
			};
		int iDataSize = PIFInfo_getSize( pInfo );
		PIFInfo_delete( pInfo );
		char *pInData = PI_NEW( char[iDataSize] );
		if ( pInData )
			{
			ifs.read( pInData, iDataSize );
			};
		if ( !pInData || !*pInData )
			{
			CONFIG_ERR( pObject, "StubIO: could not allocate output \
buffer or datafile is empty" );
			return 0;
			};
		
		return PI_NEW( IOResponseRecord( pInData, iDataSize ) );
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class StubIO
{
private:
	PIObject *pObject;
	StubIO *pIOObject;	/* prototype IO object from which this was created */
	int iAcceptInterval;/* from config. 0 = no wait, -1 = infinite wait */ 
	int iSendInterval;	/* from config. 0 = no wait, -1 = infinite wait */ 
	DblList lRecords;	/* response records */
	DblList lHTTPEquivalents;
						/* rfc822 headers for DB */
	int iErrorAfter;	/* error after */
	const IOResponseRecord *pIORecord;
						/* current record of child IO */
	int iNextRecord;	/* prototype objects next record index for child IO */
	int iPos;			/* next position on pInData */
	int bOK;
	PIDB *pSaveDB;

	/* --- parameters --- */
	int Parameter( const char *pVar, const char *pVal, const char *pWhere )
		{
		assert( pVar );
		assert( pVal );
		PIOStrStream os;
		os << pWhere << "StubIO: ";
		if ( !PIUtil_stricmp( pVar, KEY_CONF_DATAFILE ) )
			{
			IOResponseRecord *pRec = IOResponseRecord::MakeResponseRecord(
				pObject, pVal );
			if ( !pRec )
				{ return 0; };
			lRecords.Append( (DblList::type)pRec );
			}
		else if ( !PIUtil_stricmp( pVar, KEY_CONF_ERRORAFTER ) )
			{
			iErrorAfter = atoi( pVal );
			}
		else if ( !PIUtil_stricmp( pVar, KEY_CONF_HTTPEQUIV ) )
			{
            RFC822Header *pHeader = PI_NEW( RFC822Header( pVal,
				PIDBTYPE_STRING ) );
            if ( !pHeader || !pHeader->IsOK() )
                {
                os << "Bad HTTPEquiv expression, '" << pVal << "'" << ends;
                CONFIG_ERR( pObject, os.str() );
                PI_DELETE( pHeader );
                return 0;
                }
            else
                {
                lHTTPEquivalents.Append( (DblList::type)pHeader );
                };
			}
		else if ( !PIUtil_stricmp( pVar, KEY_CONF_ACCEPTINTERVAL ) )
			{
			iAcceptInterval = atoi( pVal );
			}
		else if ( !PIUtil_stricmp( pVar, KEY_CONF_SENDINTERVAL ) )
			{
			iSendInterval = atoi( pVal );
			}
		else
			{
			os << "Unknown parameter '" << pVar << "'." << ends;
			CONFIG_ERR( pObject, os.str() );
			return 0;
			};

		/* --- OK --- */
		return 1;
		};
	static int ParameterFn( void *pData, const char *pVar, const char *pVal,
		const char *pWhere )
		{
		assert( pData );
		return ((StubIO *)pData)->Parameter( pVar, pVal, pWhere );
		};

	/* --- forbid copy constructor --- */
	StubIO( const StubIO & )
		{ assert( 0 ); };

public:
	StubIO( PIObject *pTheObject, int iArgc, const char *ppArgv[] )
	:
		pObject( pTheObject ),
		pIOObject( 0 ),
		iAcceptInterval( DEFAULT_INTERVAL ),
		iSendInterval( DEFAULT_INTERVAL ),
		lRecords(),
		lHTTPEquivalents(),
		iErrorAfter( -1 ),
		pIORecord( 0 ),
		iNextRecord( 0 ),
		iPos( 0 ),
		bOK( 0 ),
		pSaveDB( 0 )
		{
		/* --- constructor --- */

		if ( !PIObject_readParameters( pObject, iArgc, ppArgv,
			StubIO::ParameterFn, this ) )
			{ return; };

		if ( !lRecords.Size() )
			{
			CONFIG_ERR( pObject, "StubIO: no definition for \
'DataFile'.");
			return;
			};

		bOK = ( lRecords.Size()>0 );
		return;
		};

	/* --- */
	StubIO( PIObject *pTheObject, StubIO *pPrototype, int, const char *[] )
	:
		pObject( pTheObject ),
		pIOObject( pPrototype ),
		iAcceptInterval( DEFAULT_INTERVAL ),
		iSendInterval( DEFAULT_INTERVAL ),
		lRecords(),
		lHTTPEquivalents(),
		iErrorAfter( -1 ),
		pIORecord( 0 ),
		iNextRecord( 0 ),
		iPos( 0 ),
		bOK( 0 ),
		pSaveDB( 0 )
		{
		assert( pIOObject );
		if ( !pIOObject ) { return; };

		/* ---
		Make new child IO DB
		--- */
		pSaveDB = PIObject_getDB( pObject );
		PIDB *pDB = PIDB_new( 0, "DB" );
		PIObject_setDB( pObject, pDB );

		/* ---
		Check error after number
		--- */
		if ( pIOObject->iErrorAfter != -1 )
			{
			assert( (pIOObject->iErrorAfter)>=0 );
			if ( !pIOObject->iErrorAfter )
				{ return; };			/* --- Cause an error --- */
			(pIOObject->iErrorAfter)--;
			};

		/* ---
		Add the headers
		--- */
		for( DblListIterator i( pIOObject->lHTTPEquivalents );
			!i.BadIndex(); i++ )
			{
			((RFC822Header *)i.Current())->ReplaceHeader( pDB,
				PIDBTYPE_STRING );
			};

		iSendInterval = pIOObject->iSendInterval;
		iAcceptInterval = pIOObject->iAcceptInterval;
		assert( iNextRecord<pIOObject->lRecords.Size() );
		iNextRecord=pIOObject->iNextRecord++;
		pIORecord = (IOResponseRecord *)pIOObject->lRecords[iNextRecord];
		if ( pIOObject->iNextRecord>=pIOObject->lRecords.Size() )
			{ pIOObject->iNextRecord = 0; };
		bOK=1;

		/* ---
		Wait the accept timeout interval
		--- */
		if ( iAcceptInterval>0 )
			{ PIPlatform_sleep( iAcceptInterval*1000 ); };
		};

	~StubIO()
		{
		if ( !pIOObject )
			{
			/* --- prototype object --- */
			for( DblListIterator i( lRecords ); !i.BadIndex(); i++ )
				{ PI_DELETE( (IOResponseRecord *)i.Current() ); };
			for( DblListIterator j( lHTTPEquivalents ); !j.BadIndex(); j++ )
				{ PI_DELETE( (RFC822Header *)j.Current() ); };
			lRecords.Clear();
			}
		else if ( pSaveDB )
			{
			/* --- not prototype object --- */
			PIDB_delete( PIObject_getDB( pObject ) );
			PIObject_setDB( pObject, pSaveDB );
			};
		};
	
	/* --- */
	inline int Out( void *, int iLen, int *piSent, int, const char *[] )
		{
		if ( !IsOK() ) { return PIAPI_ERROR; }; 

		*piSent = iLen;
		/* -- 
		Using a random way of decreasing *piSent would simulate
		outgoing TCP fragmentation
		--- */ 

		if ( iSendInterval>0 )
			{ PIPlatform_sleep( iSendInterval*1000 ); };
		PIThread_userYield();
		return PIAPI_COMPLETED;
		};

	/* --- */
	inline int In( void *pData, int iLen, int *piReceived, int,
		const char *[] )
		{
		assert( pIORecord );
		if ( !IsOK() ) { return PIAPI_ERROR; }; 
		
		int iSend = pIORecord->iDataSize - iPos;

		/* --- 
		NOTE: using a random way to decrease iSend at this 
		point would simulate fragmentation
		--- */

		if ( iLen < iSend ) { iSend = iLen; };
		if ( iSend )
			{
			memcpy( pData, &( pIORecord->pInData[iPos] ), iSend );
			};
		*piReceived = iSend;
		iPos+=iSend;
		PIThread_userYield();
		return PIAPI_COMPLETED;
		};

	inline int IsOK()	{ return bOK; };

};

/*____________________________________________________________________________*\
 *
 Function:		 StubIO_onClassLoad
 Synopsis:
 Description:
	This is called on program startup, restart or shutdown.
\*____________________________________________________________________________*/
PUBLIC_PIAPI int StubIO_onClassLoad( PIClass_LoadAction eAction, void * )
{
	switch( eAction )
		{
		case STARTUP:
			break;

		case SHUTDOWN:
			break;
		
		default:;
		};

	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:		 StubIO_StubIO
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int StubIO_StubIO( PIObject *pObj, int iArgc,
	const char *ppArgv[] )
{
	StubIO *pStubIO = PI_NEW( StubIO( pObj, iArgc, ppArgv ) );
	if (!( pStubIO && pStubIO->IsOK() ))
		{
		PI_DELETE( pStubIO );
		return PIAPI_ERROR;
		};
	PIObject_setUserData( pObj, pStubIO );
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:		 StubIO_copyStubIO
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int StubIO_copyStubIO( PIObject *pObj, int iArgc,
	const char *ppArgv[] )
{
	StubIO *pPrototype = (StubIO *)PIObject_getUserData( pObj );
	StubIO *pStubIO = PI_NEW( StubIO( pObj, pPrototype, iArgc, ppArgv ) );
	if (!( pStubIO && pStubIO->IsOK() ))
		{
		PIObject_setUserData( pObj, 0 );
		PI_DELETE( pStubIO );
		return PIAPI_ABORT;
		};
	PIObject_setUserData( pObj, pStubIO );
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:		StubIO_xStubIO
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int StubIO_xStubIO( PIObject *pObj, int,
	const char *[] )
{
	assert( pObj );
	StubIO *pStubIO = (StubIO *)PIObject_getUserData( pObj );
	if ( pStubIO ) { PI_DELETE( pStubIO ); };
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:   	StubIO_out
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int StubIO_out( PIObject *pObj, void *pData, int iLen,
		int *piSent, int iArgc, const char *ppArgv[] )
{
	assert( pObj && PIObject_getUserData( pObj ) );
	return ((StubIO *)PIObject_getUserData( pObj ) )->Out( pData, iLen,
		piSent, iArgc, ppArgv );
}

/*____________________________________________________________________________*\
 *
 Function:   	StubIO_in
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int StubIO_in( PIObject *pObj, void *pData, int iLen,
		int *piReceived, int iArgc, const char *ppArgv[] )
{
	assert( pObj && PIObject_getUserData( pObj ) );
	return ((StubIO *)PIObject_getUserData( pObj ) )->In( pData, iLen,
		piReceived, iArgc, ppArgv );
}

/* ---
	Class configuration specification 
--- */
#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name StubIOClass
		Type IO
		Library IO
		OnClassLoad StubIO_onClassLoad
		Constructor StubIO_StubIO
		CopyConstructor StubIO_copyStubIO
		Destructor StubIO_xStubIO
		Out StubIO_out
		In StubIO_in
	</Class>

	<Object>
		Name StubIO
		Class StubIOClass
		AcceptInterval 0
		SendInterval 0
	</Object>

/*___+++CNF_END+++___*/
#endif

