/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: SThread.cpp$
 * $Date: Sun Aug 10 06:35:59 1997$
 *
 Description:
	Single threaded IO server. Handles only one request at a time.
\*____________________________________________________________________________*/
//$SourceTop:$

#include <assert.h>
#include <iostream.h>
#include <stdlib.h>
#include "IOSrvBse.h"
#include "Pi2API.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define CONFIG_ERR(this, msg)   \
	{ PILog_addMessage( PIObject_getDB(this), \
	PIObject_getConfigurationDB(this), PILOG_ERROR, (msg) ); }
#define CONFIG_WARN(this, msg)   \
	{ PILog_addMessage( PIObject_getDB(this), \
	PIObject_getConfigurationDB(this), PILOG_WARNING, (msg) ); }

/*____________________________________________________________________________*\
 *
 Description:
	Documentation.
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this server
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	SingleThreadedIOServer

Description:
	This object performs single threaded dispatch of IO requests. On
	execution it waits for a new IO request, and completes processing
	of the request before accepting a new request for processing.
	<P>
	This dispatching object initializes an IO object for accepting IO 
	requests, and a logic object for handling those requests.

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>IOObject
<TD>+
<TD>A Pi3 IO object name
<TD>A Pi3 object
<TD>IOObject "ServerTCPIPObject"

<TR>
<TD>LogicObject
<TD>+
<TD>A Pi3 Logic object name
<TD>A Pi3 object
<TD>LogicObject "HandleRequestObject"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>
	IOObject
</H5>
Specifies the IO object used to accept new IO requests. The IO object is
created during construction. During execution of this object PIObject_copy()
is invoked to create a copy IO object for each new request. When processing
of the request is completed, the new IO object is destroyed and the server
waits for another IO request by again calling PIObject_copy() on the main
IO object.
The IO object can be an object which provides an IO service (such as TCP/IP,
named Pipe or shared memory IO) or an IO filter which performs translation
on such an IO object (such as encryption or monitoring).

<H5>
	LogicObject
</H5>
Specifies the logic object used to handle IO requests. The object is 
created during construction and its <I>execute</I> method invoked
(using PILogic_execute()) for each IO request accepted. The IO object
for the request is passed as an argument to this logic object.

Returns:
	PIAPI_COMPLETED on success.
	PIAPI_ERROR and PIAPI_ABORT respectively for generic and severe
	error conditions.

Example:
	<PRE>
	&ltObject&gt
		Name SingleThreadedIOServer
		Class SingleThreadedIOServerClass
		IOObject ServerTCPIPIO
		LogicObject HandleHTTP
	&lt/Object&gt
	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class SingleThreadedIOServer : public IOServerBase
{
private:
public:
	SingleThreadedIOServer( const PIObject *pTheObject )
	:	IOServerBase( pTheObject )
		{
		if ( !bOK ) { return;	/* Base class initialization failed */ };
		bOK = 1;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int SingleThreadedIOServer_onClassLoad( PIClass_LoadAction eLoad,
	int, const char *[] )
{
	switch( eLoad )
		{
		case STARTUP:
		default:;
		};

	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int SingleThreadedIOServer_constructor( PIObject *pObj, int,
	const char *[] )
{
	IOServerBase *pServer = new SingleThreadedIOServer( pObj );
	if (!( pServer && pServer->IsOK() ))
		{
		delete pServer;
		return PIAPI_ERROR;
		};
	if ( PIObject_setUserData( pObj, pServer ) )
		{
		CONFIG_ERR( pObj,
		"SingleThreadedIOServer_constructor: PIObject_setUserData() failed" );
		return PIAPI_ERROR;
		};
	return PIAPI_COMPLETED;
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name SingleThreadedIOServerClass
		Type LogicExtension
		Library Server
		OnClassLoad SingleThreadedIOServer_onClassLoad
		Constructor SingleThreadedIOServer_constructor
		CopyConstructor SingleThreadedIOServer_copyConstructor
		Destructor IOServerBase_destructor
		Execute IOServerBase_execute
	</Class>

	<Object>
		Name SingleThreadedIOServer
		Class SingleThreadedIOServerClass
	</Object>

/*___+++CNF_END+++___*/
#endif

